# -*- coding: UTF-8 -*-

import filecmp
import fnmatch
import glob
import os
import shutil
import sys
import time
import zipfile
from datetime import timedelta, datetime
from multiprocessing import RawValue, Lock
from operator import itemgetter

sys.path.append(os.path.dirname(sys.path[0]))


class Counter(object):
    def __init__(self, value=0):
        # RawValue because we don't need it to create a Lock:
        self.val = RawValue('i', value)
        self.lock = Lock()

    def increment(self):
        with self.lock:
            self.val.value += 1

    def incAndGet(self):
        with self.lock:
            self.val.value += 1
            return self.val.value

    def value(self):
        with self.lock:
            return self.val.value


counter = Counter(0)


def copyMoveFiles(arrayPatterns, destDirectory, move=False):
    files = []
    for f in arrayPatterns:
        for file in os.listdir(os.path.dirname(f)):
            if fnmatch.fnmatch(file.lower(), os.path.basename(f).lower()):
                if move and not destDirectory:
                    os.remove(os.path.join(os.path.dirname(f), file))
                elif not os.path.exists(os.path.join(destDirectory, file)):
                    if move:
                        shutil.move(os.path.join(os.path.dirname(f), file), destDirectory)
                    else:
                        shutil.copy(os.path.join(os.path.dirname(f), file), destDirectory)
                files.append("%s/%s" % (destDirectory, file))
    return files


def copyFiles(arrayPatterns, destDirectory):
    return copyMoveFiles(arrayPatterns, destDirectory, False)


def moveFiles(arrayPatterns, destDirectory):
    return copyMoveFiles(arrayPatterns, destDirectory, True)


def removeFiles(arrayPatterns):
    copyMoveFiles(arrayPatterns, None, True)


def sort_files_by_last_modified(files):
    fileData = {}
    for fname in files:
        fileData[fname] = os.stat(fname).st_mtime
    fileData = sorted(fileData.items(), key=itemgetter(1))
    return fileData


def archive(log, srcDir, dstDir, maxCountFiles=30, actualCountDays=30, removeSRC=True, fileName=None, loginfo=True):
    if not srcDir or not dstDir:
        log.warn(u'Невозможно сделать бэкап, т.к. указаны некорректные данные: srcDir = %s, backupDir = %s' % (str(srcDir), str(dstDir)))
        return
    if loginfo:
        log.info(u'Делаем бэкап: %s -> %s' % (str(srcDir), str(dstDir)))
    id = os.path.basename(os.path.dirname(srcDir))
    root = os.path.join(dstDir, id)
    ext = '.zip'
    # В многопоточном soap-режиме может возникнуть ситуация, что несколько потоков одновременно начнет сжимать в один файл.
    # Тут может получится так, что часть архива не будет читаться, или сервис вообще вылетит с исключением. Для нивелирования
    # каждый файл архива создается с уникальным значением счетчика
    backup = root + str(time.time()).replace('.', '') + str(counter.incAndGet()) + ext
    if not os.path.exists(dstDir):  # проверка не существует ли такая директория
        os.makedirs(dstDir)
    gz = zipfile.ZipFile(backup, 'a', 8, True)
    try:
        for file in os.listdir(srcDir):
            if not fileName or file == fileName:
                gz.write(os.path.join(srcDir, file), file, 8)
                if removeSRC:
                    os.remove(os.path.join(srcDir, file))
        if removeSRC and not fileName:
            os.rmdir(srcDir)
    except zipfile.BadZipfile as e:
        log.error(u'Не удалось создать архив %s. %s' % (backup, str(e)))
        return
    except Exception as e:
        log.error(u'Ошибка: %s' % str(e))
        return
    finally:
        gz.close()

    if loginfo:
        log.info(u'Archive = ' + backup)

    if maxCountFiles > 0:
        files = glob.glob('%s/%s*.zip' % (dstDir, id))
        sorted_files = sort_files_by_last_modified(files)
        deleteCount = len(sorted_files) - maxCountFiles
        for x in range(0, deleteCount):
            log.debug(u"Удаляем устаревший архив %s" % str(sorted_files[x][0]))
            os.remove(sorted_files[x][0])
    if actualCountDays > 0:
        files = glob.glob('%s/%s*.zip' % (dstDir, id))
        for fname in files:
            actualDate = datetime.now() - timedelta(days=actualCountDays)
            if datetime.fromtimestamp(os.stat(fname).st_mtime) < actualDate:
                log.debug(u"Удаляем устаревший архив %s" % str(fname))
                os.remove(fname)


def getFilesByMask(directory, arrayPatterns):
    files = []
    for pattern in arrayPatterns:
        for file in os.listdir(directory):
            if fnmatch.fnmatch(file.lower(), os.path.basename(pattern).lower()):
                files.append(os.path.join(directory, file))
    return files


def safeCopyFile(log, oldFilePath, newFilePath, rewrite=False):
    log.info(u"Копирование файла %s в %s" % (oldFilePath, newFilePath))
    suffix = '_FILE_PART'
    try:
        if not os.path.exists(os.path.dirname(newFilePath)):
            os.makedirs(os.path.dirname(newFilePath))
        if os.path.exists(newFilePath + suffix):
            os.remove(newFilePath + suffix)
#        if not os.path.exists(os.path.dirname(newFilePath)):
#            os.makedirs(os.path.dirname(newFilePath))
        shutil.copyfile(oldFilePath, newFilePath + suffix)
        attempts = 0
        while not filecmp.cmp(oldFilePath, newFilePath + suffix) and attempts < 3:
            os.remove(newFilePath + suffix)
            shutil.copyfile(oldFilePath, newFilePath + suffix)
            attempts += 1
        if not filecmp.cmp(oldFilePath, newFilePath + suffix):
            log.error('Can not make good copy.')
            return False
        if os.path.exists(newFilePath) and rewrite:
            log.info('Rewrite = %s' % rewrite)
            os.remove(newFilePath)
        if not os.path.exists(newFilePath):
            os.rename(newFilePath + suffix, newFilePath)
        else:
            os.remove(newFilePath + suffix)
            return False
        shutil.copymode(oldFilePath, newFilePath) # копирование прав доступа
#            os.remove(oldFilePath)
    except Exception as e:
        log.info(str(e))
        return False
    return True


def safeMoveDirectory(log, oldDirPath, newDirPath, rewrite = False, backupDir = None, maxCountFiles=30, actualCountDays=30):
    log.info(u"Перемещение файлов из директории %s в %s" %(oldDirPath, newDirPath))
    removeFiles = []
    copiedFiles = []
    allFileProcessed = True
    for file in os.listdir(oldDirPath):
        if file != 'ready' and file != 'readyOnline' and os.path.isfile(os.path.join(oldDirPath, file)):
            remoteName = os.path.join(newDirPath, file)
            if not safeCopyFile(log, os.path.join(oldDirPath, file), remoteName, rewrite = rewrite):
                log.info(u"Can not copy %s from workDirectory to %s" % (file, newDirPath))
                allFileProcessed = False
            else:
                removeFiles.append(os.path.join(oldDirPath, file))
                copiedFiles.append(remoteName)
    if len(removeFiles) > 0:
        if backupDir:
            archive(log, oldDirPath, backupDir, maxCountFiles=int(maxCountFiles), actualCountDays=int(actualCountDays), removeSRC=False)
    else:
        log.info('Sales is empty.')
    for f in removeFiles:
        os.remove(f)
    return allFileProcessed, copiedFiles


def simpleCopyFiles(fromDir, toDir, log):
    log.info(u"Копирование файлов из директории %s в %s" %(fromDir, toDir))
    allFileProcessed = True
    copiedFiles = []
    if not os.path.isdir(toDir):
        os.makedirs(toDir)
    path, dirs, files = next(os.walk(fromDir))
    for name in files:
        log.info(u"Копирование файла %s" % (os.path.join(fromDir, name)))
        try:
            shutil.copy(os.path.join(fromDir, name), os.path.join(toDir, name))
            copiedFiles.append(os.path.join(toDir, name))
        except Exception as e:
            log.info(str(u"Ошибка копирования файла: %s") % e)
            allFileProcessed = False

    return allFileProcessed, copiedFiles


def checkFilesInRemoteDirectory(rewrite, workDirectory, remoteDirectory):
    remoteDir = remoteDirectory
    if rewrite:
        return True
    if os.path.exists(remoteDir):
        for f in os.listdir(workDirectory):
            remoteFile = os.path.join(remoteDir, f)
            if os.path.isfile(remoteFile):
                with open(remoteFile, "r") as ff:
                    if ff.readline(1) == '@':
                        return True
    return False


def removeFilesInDirectory(log, directory):
    log.info(u"Удаление файлов из директории %s" % directory)
    path, dirs, files = next(os.walk(directory))
    for name in files:
        log.info(u"Удаление файла %s" % os.path.join(directory, name))
        try:
            os.remove(os.path.join(directory, name))
        except Exception as e:
            log.error(str(u"Ошибка удаления файла: %s") % e)
            return False
    return True


def copyFilesByFlags(flags, destDirectory):
    files = []
    for flag in flags:
        for file in os.listdir(os.path.dirname(flag)):
            fname, fext = os.path.splitext(flag)
            dname, dext = os.path.splitext(file)
            if os.path.basename(fname).lower() == dname.lower() and fext.lower() != dext.lower():
                shutil.copy(os.path.join(os.path.dirname(flag), file), destDirectory)
                files.append("%s/%s" % (destDirectory, file))
    return files


def removeFilesByFlags(flags):
    for flag in flags:
        for file in os.listdir(os.path.dirname(flag)):
            fname, fext = os.path.splitext(flag)
            dname, dext = os.path.splitext(file)
            if os.path.basename(fname).lower() == dname.lower() and fext.lower() != dext.lower():
                os.remove(os.path.join(os.path.dirname(flag), file))


def addFilePermissions(log, fileName, permissions):
    if os.path.exists(fileName):
        log.info(u'Изменение прав доступа для файла %s, %s' % (fileName, oct(permissions)))
        try:
            mode = os.stat(fileName).st_mode
            mode |= permissions
            os.chmod(fileName, permissions)
        except Exception as e:
            log.warn(repr(e))
