# -*- coding: UTF-8 -*-

import os
from common import filesmodify, logger, macros
import datetime
import codecs
import sys


class AtolBasic(object):
    """
    @brief Базовый класс для формата Атол
    Базовая реализация для конвертера выгрузки Атол.
    Общий для конвертора Atol и Atol-Astor.
    """
    def __init__(self):
        self.id = 0
        self.currentConverter = None
        self.typesInKKM = {
            1: 0,  # продажа
            2: 1,  # возврат
            3: 4,  # внесение
            4: 5,  # выем
            7: 2,  # аннулирование чека продажи
            8: 2,  # аннулирование возврата
            16: 16,  # инвентаризация
            25: 1,  # возврат по чеку продажи
        }
        self.saleFiles = set()

    def parseFlag(self, flag):
        shifts = None
        dates = None
        with open(flag, 'r') as f:
            if f.readline().strip() == '###':
                str = f.readline().strip()
                if len(str.split('-')) == 2:
                    if str.find('.') == -1:
                        shifts = str.replace('-', '@')
                    else:
                        fromDate = str.split('-')[0].split('.')
                        toDate = str.split('-')[1].split('.')
                        dates = '%s-%s-%sT00:00:00.0@%s-%s-%sT23:59:59.99' % (
                        fromDate[2], fromDate[1], fromDate[0], toDate[2], toDate[1], toDate[0])
        return shifts, dates

    def openFile(self, log, cash, node, workDirectory, shiftNum, currentSales, previousFile=None):
        if currentSales:
            return self.openOnlineUnloadFile(log, cash, node, workDirectory, shiftNum, previousFile=None)
        return self.openUnloadFile(log, cash, node, workDirectory, shiftNum, previousFile=None)

    def openOnlineUnloadFile(self, log, cash, node, workDirectory, shiftNum, previousFile=None):
        output = None
        currentPath = macros.applyMacros(self.currentConverter.onlineSalePath, log, shopcode=node.id, cashcode=cash.code, shiftnum=shiftNum)

        salePath = workDirectory + '/' + os.path.basename(currentPath)
        if os.path.isfile(salePath):
            log.info(u"Файл уже существует: " + salePath)
            output = codecs.open(salePath, 'a', self.currentConverter.charset)

        if not output:
            output = codecs.open(salePath, 'w', self.currentConverter.charset)
            log.info(u"Файл создан: " + salePath)

        return output

    def openUnloadFile(self, log, cash, node, workDirectory, shiftNum, previousFile=None):
        output = None
        currentPath = macros.applyMacros(self.currentConverter.path, log, shopcode=node.id, cashcode=cash.code, shiftnum=shiftNum)

        salePath = workDirectory + '/' + os.path.basename(currentPath)
        if os.path.isfile(salePath):
            log.info(u"Файл уже существует: " + salePath)

            f = codecs.open(salePath, 'r', self.currentConverter.charset)
            if f.readline(1) == '#':
                if not self.currentConverter.rewriteDestFile or previousFile == salePath:
                    output = codecs.open(salePath, 'a', self.currentConverter.charset)
            try:
                f.next()  # первая строка (# или @)
                f.next()  # вторая строка с номером кассы
                f.next()  # третья строка с номером смены
            except StopIteration:  # если нет одной из строк, то работаем с файлом так же, как если бы его не было
                if output:
                    output.close()
            f.close()

        if not output:
            output = codecs.open(salePath, 'w', self.currentConverter.charset)
            output.write("#\r\n")
            cashCode = str(cash.code)
            if sys.version_info.major == 2:
                cashCode = unicode(cashCode, 'utf-8')
            output.write(cashCode + "\r\n")
            output.write(shiftNum + "\r\n")
            log.info(u"Файл создан: " + salePath)

        return output

    def createServiceFile(self, log, cash, node, workDirectory, shiftNum, id):
        currentPath = macros.applyMacros(self.currentConverter.onlineSalePath, log, shopcode=node.id, cashcode=cash.code, shiftnum=shiftNum)
        serviceFilePath = workDirectory + '/' + "_" + os.path.basename(currentPath)
        output = codecs.open(serviceFilePath, 'w', self.currentConverter.charset)
        cashCode = str(cash.code)
        if sys.version_info.major == 2:
            cashCode = unicode(cashCode, 'utf-8')
        output.write(cashCode + "," + str(id))
        output.close()

    def run(self, salesFunc, securityLogsFunc, shiftsFunc, cash, node, workDirectory, currentSales):
        log = logger.getUnloadLogger("Node", node.id)
        log.info(u"Начинаем выгрузку. Касса №" + str(cash.code))
        self.currentConverter = node.currentConverter
        output = None
        firstRound = True       # Флаг. Указывает на то, чтобы писать или не писать заголовок отчета в файл.

        shifts = {}
        for sh in shiftsFunc():
            shifts[sh['shift']] = sh
        shift = None
        shopCode = None

        for check in salesFunc():
            if firstRound:
                output = self.openFile(log, cash, node, workDirectory, str(check['shift']), currentSales)
                firstRound = False

            if shift != check['shift']:
                if shift in shifts:  # если предыдущая смена обработана
                    self.workShiftClosing(output, shifts[shift], shopCode=shifts[shift]['shopcode'])
                    previousFile = output.name
                    output.close()
                    output = self.openFile(log, cash, node, workDirectory, str(check['shift']), currentSales, previousFile)
                shift = check['shift']
                shopCode = shifts[shift]['shopcode']
                self.id = 100000 * check['shift']
                self.workShiftOpening(output, shifts[shift], shopCode)

            if check['docType'] == 13:
                continue

            opTypeInKKM = self.typesInKKM[check['docType']] if check['docType'] in self.typesInKKM else 0
            self.documentOpening(output, check, opTypeInKKM, shopCode)

            listDiscForCheck = []
            positions = check['inventPositions'] + check['stornoPositions']
            
            for pos in sorted(positions, key=lambda pos: pos['posNum']):
                self.positionsAdd(output, pos, check["cashCode"], check["docNum"], opTypeInKKM, check['shift'], shopCode)
                if 'timeStorno' in pos:
                    self.stornoPositionsRemove(output, pos, check["cashCode"], check["docNum"], opTypeInKKM, check['shift'], shopCode)
                else:
                    res = self.discounts(output, pos, self.currentConverter.discountsDetail, check['docType'], check['cashCode'], check['docNum'], check['baseSum'], 
                                    opTypeInKKM, check['shift'], shopCode)
                    if len(res):
                        listDiscForCheck += res

            if len(listDiscForCheck):
                self.discountCheck(output, listDiscForCheck, check['docNum'], check['docType'], opTypeInKKM, check['shift'], shopCode, check["cashCode"])

            self.documentPayment(output, check["userCode"], check["cashCode"], check["docNum"], check['moneyPositions'], check['inventPositions'], opTypeInKKM, check['shift'], shopCode)

            if check['docType'] in [1, 2, 7, 8, 25]:  # Чек продажи/возврата
                if check['closed'] in [1, 2]:
                    self.documentClosing(output, check, opTypeInKKM, shopCode)
            elif check['docType'] in [3, 4]:  # Внесения/изъятие денег
                self.moneyOperations(output, check, opTypeInKKM, shopCode)
                
        if shift in shifts:
            self.workShiftClosing(output, shifts[shift], shifts[shift]['shopcode'])

        if output:
            output.close()
            statusFileName = "ready"
            statusDir = workDirectory
            if currentSales:
                #createServiceFile(log, cash, node, workDirectory, shift, id)
                statusFileName += "Online"
                statusDir = os.path.dirname(workDirectory)

            converterSalesReadyFile = open('%s/%s' % (statusDir, statusFileName), 'w')
            converterSalesReadyFile.close()
            salesReadyFile = open('%s/%s' % (os.path.dirname(os.path.dirname(statusDir)), statusFileName), 'w')
            salesReadyFile.close()
            self.saleFiles = set((output.name, converterSalesReadyFile.name, salesReadyFile.name))

        log.info(u"Выгружены следующие номера смен: %s" % shifts.keys())
        log.info(u"Закончили выгрузку")
        
    def discountCheck(self, file, listCheckDisc, docNum, docType, opTypeInKKM, shiftNum, shopCode, cashCode):
        totalDiscount = {}
        for disc in listCheckDisc:        
            if disc['discNumber'] not in totalDiscount:
                totalDiscount[disc['discNumber']] = disc
                totalDiscount[disc['discNumber']]['DISCOUNTSUM'] = disc['discSum']
                totalDiscount[disc['discNumber']]['MAXDISCSIZE'] = disc['discSize']
            else:
                totalDiscount[disc['discNumber']]['DISCOUNTSUM'] += disc['discSum']
                totalDiscount[disc['discNumber']]['MAXDISCSIZE'] = max(totalDiscount[disc['discNumber']]['MAXDISCSIZE'], disc['discSize'])
        
        for key in totalDiscount:
            prm11 = prm12 = abs(totalDiscount[key]['DISCOUNTSUM'])        
                
            if docType != 1:
                prm12 *= (-1)
                
            typeTrans = 35
            
            if totalDiscount[key]['discType'] == 'percent':
                prm11 = totalDiscount[key]['MAXDISCSIZE']
                typeTrans = 37        
            if totalDiscount[key]['DISCOUNTSUM'] < 0:
                typeTrans += 1
            
            cardNumber = ''
            if totalDiscount[key]['discMode'] == 'byCard':
                cardNumber = totalDiscount[key]['cardNumber']
            
            discType = 2 #автоматическая
            discScheme = totalDiscount[key]['discScheme']
            discCode = totalDiscount[key]['discCode']
            pos26 = '%s/%s/%s' % (cashCode, shiftNum, docNum)
            
            self.insertData(file, totalDiscount[key]['posTime'], typeTrans, totalDiscount[key]['cashcode'], docNum, totalDiscount[key]['userCode'], cardNumber, '', 
                    discType, prm11, prm12, opTypeInKKM, shiftNum, discScheme, discCode, '', '', '', '', '', '', '', '', '', pos26, shopCode, '', '', '')
            
    def discountPosition(self, file, disc, cashCode, docNum, docType, position, opTypeInKKM, shiftNum, shopCode, isPositionDiscount=True):
        typeTrans = 15
        discType = 2 #автоматическая
        discRate = disc['discSize']
        discSum = -abs(disc['discSum'])
        discScheme = disc['discScheme']
        discCode = disc['discCode']
        pos26 = '%s/%s/%s' % (cashCode, shiftNum, docNum)
        cardNumber = ''
        
        if disc['discMode'] == 'byCard':
            cardNumber = disc['cardNumber']
        
        if disc['discType'] == 'percent':
            typeTrans = 17
        if disc['discSum'] < 0: 
            typeTrans += 1
        
        if disc['discType'] in ['absolute', 'index']:
            discRate = disc['discSum']
                    
        if docType == 1:
            discSum *= (-1)

        if position['opCode'] in [52, 56, 60]:
            typeTrans -= 10

        if not isPositionDiscount:
            typeTrans += 70
        self.insertDiscountData(file, disc['posTime'], typeTrans, cashCode, docNum, disc['userCode'], cardNumber, discType, discRate, discSum, opTypeInKKM, shiftNum, 
                discScheme, discCode, pos26, shopCode, position)

    def insertDiscountData(self, file, timeTrans, typeTrans, cashCode, docNum, userCode, cardNumber, discType, discRate, discSum, opTypeInKKM, shiftNum, discScheme, discCode, pos26, shopCode, position):
        """
        @brief Добавление данных о скидке
        """
        self.insertData(file, timeTrans, typeTrans, cashCode, docNum, userCode, cardNumber, '', discType, discRate, discSum, opTypeInKKM, shiftNum, 
            discScheme, discCode, '', '', '', '', '', '', '', '', '', pos26, shopCode, '', '', '')
        
    def discounts(self, file, position, discountsDetail, docType, cashCode, docNum, docBaseSum, opTypeInKKM, shiftNum, shopCode):   
        listDiscForCheck = []
        
        for disc in position['discountPositions']:
            if disc['isPositionDiscount']:
                self.discountPosition(file, disc, cashCode, docNum, docType, position, opTypeInKKM, shiftNum, shopCode)
            else:
                if discountsDetail:
                    self.discountPosition(file, disc, cashCode, docNum, docType, position, opTypeInKKM, shiftNum, shopCode, isPositionDiscount = False)
                else:                
                    listDiscForCheck.append(disc)                
        return listDiscForCheck

    def workShiftOpening(self, file, shift, shopCode):
        docNum = 0
        self.insertData(file, shift['timeBeg'], 62, shift['cashCode'], docNum, shift['userCode'], '', '', '', '', '', 8, shift['shift'], '', '', '', '', '', '', 
                '', '', '', '', '', '', shopCode, '', '', '')

    def workShiftClosing(self, file, shift, shopCode):
        if shift['timeEnd']:
            docNum = 1
            if shift['checkNum2']:
                docNum = shift['checkNum2'] + 1

            for kkm in shift['kkms']:
                self.insertData(file, shift['timeEnd'], 63, shift['cashCode'], docNum, shift['userCode'], '', '',
                                kkm['sumGain'], kkm['sumCashEnd'], kkm['sumSale'], 9, shift['shift'], '', '',
                                kkm['kkmNum'], '', '', '',
                                '', '', '', '', '', '', shopCode, '', '', '')
                self.insertData(file, shift['timeEnd'], 61, shift['cashCode'], docNum, shift['userCode'], '', '',
                                kkm['sumGain'],  '', kkm['sumSale'], 10, shift['shift'], '', '', kkm['kkmNum'], '', '',
                                '', '', '', '', '', '', '', shopCode, '', '', '')

            if not shift['kkms']:
                self.insertData(file, shift['timeEnd'], 63, shift['cashCode'], docNum, shift['userCode'], '', '',
                                shift['sumGain'], '', shift['sumSale'], 9, shift['shift'],
                                '', '', '', '', '', '', '', '', '', '', '', '', shopCode, '', '', '')
                self.insertData(file, shift['timeEnd'], 61, shift['cashCode'], docNum, shift['userCode'], '', '',
                                shift['sumGain'], '', shift['sumSale'], 10, shift['shift'],
                                '', '', '', '', '', '', '', '', '', '', '', '', shopCode, '', '', '')

    def documentOpening(self, file, check, opTypeInKKM, shopCode):
        pos26 = '%s/%s/%s' % (check['cashCode'], check['shift'], check['docNum'])
        self.insertData(file, check['timeBeg'], 42, check['cashCode'], check['docNum'], check['userCode'], '', '', '', '', '', opTypeInKKM, check['shift'], '', 
                '', '', '', '', '', '', '', '', '', '', pos26, shopCode, '', '', '', '')

    def documentClosing(self, file, check, opTypeInKKM, shopCode):
        pos12 = check['docSum']
        typeTrans = 55
        pos26 = '%s/%s/%s' % (check['cashCode'], check['shift'], check['docNum'])
        positionCount = len(check['inventPositions'])
        aspect = ''
        if check['closed'] == 2: # аннулирован
            pos12 *= (-1)
            positionCount *= (-1)
            typeTrans = 56
            aspect = 0
        self.insertData(file, check['timeEnd'], typeTrans, check['cashCode'], check['docNum'], check['userCode'], '', '', aspect, positionCount, pos12, opTypeInKKM, 
                check['shift'], '', '', '', '', '', '', '', '', '', '', '', pos26, shopCode, '', '', '', '')

    def documentPayment(self, file, userCode, cashCode, docNum, payments, positions, opTypeInKKM, shiftNum, shopCode):
        paytime = datetime.datetime.max
        pos26 = '%s/%s/%s' % (cashCode, shiftNum, docNum)
        sum = 0

        depts = {}
        for position in positions:
            depts[position['deptCode']] = depts.get(position['deptCode'], 0) + position['posSum']

        matchingPayments = []
        for pay in payments:
            if pay['opCode'] in [70, 72, 74, 76, 100, 101, 102, 103]:
                paytime = min(paytime, pay['posTime'])
                sum = abs(pay['sumB']) if pay['opCode'] in [70, 76, 101, 102] else -abs(pay['sumB'])
                matchingPayments.append({'paytime' : paytime, 'sum' : sum, 'valCode' : pay.get('valCode', 0)})

        payment = 0
        typeTrans = 40
        if len(depts) > 1:
            typeTrans = 43
            for deptCode, deptSum in sorted(depts.items()):
                for i in range(payment, len(matchingPayments)):
                    if abs(matchingPayments[i]['sum']) > deptSum:
                        if matchingPayments[i]['sum'] < 0:
                            matchingPayments[i]['sum'] += deptSum
                        else:
                            matchingPayments[i]['sum'] -= deptSum
                        sum = deptSum
                    else:
                        deptSum -= abs(matchingPayments[i]['sum'])
                        sum = matchingPayments[i]['sum']
                        payment = i + 1

                    self.insertData(file, matchingPayments[i]['paytime'], typeTrans, cashCode, docNum, userCode, '', matchingPayments[i]['valCode'], '', sum, sum,
                            opTypeInKKM, shiftNum, '', '', deptCode, '', '', '', '', '', '', '', '', pos26, shopCode, '', '', '')

                    if sum == deptSum or deptSum == 0:
                        break

        for i in range(payment, len(matchingPayments)):
            self.insertData(file, matchingPayments[i]['paytime'], typeTrans, cashCode, docNum, userCode, '', matchingPayments[i]['valCode'], '', matchingPayments[i]['sum'],
                     matchingPayments[i]['sum'], opTypeInKKM, shiftNum, '', '', list(depts.keys())[-1], '', '', '', '', '', '', '', '', pos26, shopCode, '', '', '')
        
    def moneyOperations(self, file, check, opTypeInKKM, shopCode):
        userCode = check['userCode']
        typeTrans = 51
        if check['docType'] == 3:
            typeTrans = 50
        pos26 = '%s/%s/%s' % (check['cashCode'], check['shift'], check['docNum'])
        self.insertData(file, check['timeEnd'], typeTrans, check['cashCode'], check['docNum'], userCode, 0, 0, 0, 0, check['docSum'], opTypeInKKM, check['shift'], '', 
                '', '', '', '', '', '', '', '', '', '', pos26, shopCode, '', '', '', '')
        
    def stornoPositionsRemove(self, file, invent, cashCode, docNum, opTypeInKKM, shiftNum, shopCode):
        self.positionsAdd(file, invent, cashCode, docNum, opTypeInKKM, shiftNum, shopCode, storno=True)
        
    def positionsAdd(self, file, invent, cashCode, docNum, opTypeInKKM, shiftNum, shopCode, storno=False):
        typeTrans = 11
        taxTypeTrans = 14
        timeTrans = ''

        code = invent['inventCode']
        quant = abs(invent['quant'])
        pos12 = abs(invent['modSum'])
        pos15 = float(invent['posSum']/quant)
        pos16 = abs(invent['posSum'])
        pos20 = abs(invent['baseSum'])
        pos26 = '%s/%s/%s' % (cashCode, shiftNum, docNum)

        if not invent['opCode'] in [50, 52, 63, 1000, 1004]:
            quant *= (-1)
            pos12 *= (-1)
            pos16 *= (-1)
            pos20 *= (-1)
        
        if storno:
            typeTrans = 12
            timeTrans = invent['timeStorno']
            quant *= (-1)
            pos12 *= (-1)
            pos16 *= (-1)
            pos20 *= (-1)
        else:
            timeTrans = invent['posTime']
            
        if invent['opCode'] in [52, 56, 60]:
            code = ''
            typeTrans = 2 if storno else 1
            taxTypeTrans = 4
        aspectcode = invent['aspectvaluesetcode'] if invent['aspectvaluesetcode'] else ''
        self.insertPositionData(file, timeTrans, typeTrans, cashCode, docNum, invent, code, quant, pos12, opTypeInKKM, shiftNum, pos15, pos16, pos20, pos26, shopCode, aspectcode)
        if 'timeStorno' not in invent:
            self.insertTaxData(file, timeTrans, taxTypeTrans, cashCode, docNum, invent, code, quant, pos12, opTypeInKKM, shiftNum, pos15, pos16, pos20, pos26, shopCode, aspectcode)

    def insertPositionData(self, file, timeTrans, typeTrans, cashCode, docNum, invent, code, quant, pos12, opTypeInKKM, shiftNum, pos15, pos16, pos20, pos26, shopCode, aspectcode):
        """
        @brief Добавление данных о позиции
        """
        self.insertData(file, timeTrans, typeTrans, cashCode, docNum, invent["userCode"], code, aspectcode, invent['finalPrice'], quant, pos12, 
                        opTypeInKKM, shiftNum, pos15, pos16, invent['deptCode'], '', invent['barCode'], pos20, '', '', '', '', '', pos26, shopCode, '', '', '')

    def insertTaxData(self, file, timeTrans, typeTrans, cashCode, docNum, invent, code, quant, pos12, opTypeInKKM, shiftNum, pos15, pos16, pos20, pos26, shopCode, aspectcode):
        """
        @brief Добавление данных о налоге
        """
        self.insertData(file, timeTrans, typeTrans, cashCode, docNum, invent["userCode"], code, aspectcode, invent['vatcode1'], invent['vatcode1'],
                        invent['vatsum1'], opTypeInKKM, shiftNum, pos15, pos16, invent['deptCode'], '', invent['barCode'], pos20, '', '', '', '', '', pos26, shopCode, '', '', '')
        
    def insertData(self, file, transDateTime, transType, cashCode, docNum, userCode, tab8, tab9, tab10, tab11, tab12, tab13, tab14, tab15, tab16, tab17, tab18, tab19, 
                tab20, tab21, tab22, tab23, tab24, tab25, tab26, tab27, tab28, tab29, tab30, tab31='', tab32='', tab33='', tab34='', tab35='', tab36=''):
        self.id += 1
        transNum = str(self.id)

        transDate = transDateTime.strftime('%d.%m.%Y')
        transTime = transDateTime.strftime('%H:%M:%S')
        strTrans = '%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;' \
                        % (transNum, transDate, transTime, transType, cashCode, docNum, userCode, tab8, tab9, tab10,
                        tab11, tab12, tab13, tab14, tab15, tab16, tab17, tab18, tab19, tab20, tab21, tab22, tab23,
                        tab24, tab25, tab26, tab27, tab28, tab29, tab30, tab31, tab32, tab33, tab34, tab35, tab36)

        if sys.version_info.major == 2 and not isinstance(strTrans, unicode):
            data = unicode(strTrans+'\r\n', 'utf-8')
        else:
            data = strTrans + '\r\n'

        file.write(data)

    def sendSales(self, cash, node, workDirectory, backupDir):
        log = logger.getUnloadLogger("Node", node.id)
        rewriteDestFile = filesmodify.checkFilesInRemoteDirectory(rewrite=node.currentConverter.rewriteDestFile,
                workDirectory=workDirectory, remoteDirectory=str(os.path.dirname(node.currentConverter.path)))
        return filesmodify.safeMoveDirectory(log, workDirectory, os.path.dirname(node.currentConverter.path),
                backupDir=backupDir, rewrite=rewriteDestFile, maxCountFiles=cash.maxCountFiles, actualCountDays=cash.actualCountDays)

    def sendOnlineSales(self, cash, node, workDirectory):
        log = logger.getUnloadLogger("Node", node.id)
        return filesmodify.safeMoveDirectory(log, workDirectory, os.path.dirname(node.currentConverter.onlineSalePath))

    def getSaleFiles(self):
        return self.saleFiles
